#if !defined(__RAW_H)
#define __RAW_H


// The modint raw data type, unsigned if possible
typedef double rawtype;


const rawtype MAXMODULUS = 4503599627370495.0;
const rawtype MAXRAW = 4503599627370495.0;
const int RAWBITS = 52;

// Base (must be BASEDIGIT^BASEDIGITS = BASE < all MODULI)
const rawtype BASE = 1e15;
const rawtype BASEDIGIT = 10.0;
const int BASEDIGITS = 15;
const rawtype BASEFACTORS[] = {2.0, 5.0};       // All prime factors of BASE
const int NBASEFACTORS = 2;                     // Number of prime factors of BASE

const rawtype MODULI[3] = {4222124650659841.0, 3799912185593857.0, 1899956092796929.0};         // upto n = 2^46
const rawtype PRIMITIVEROOTS[3] = {19.0, 5.0, 7.0};
const rawtype WFTAMODULI[3] = {4194068336640001.0, 2742275450880001.0, 2661620290560001.0};     // upto n = 630*5040^3
const rawtype WFTAPRIMITIVEROOTS[3] = {17.0, 19.0, 43.0};

const size_t MAXTRANSFORMLENGTH = 70368744177664UL; // Maximum transform length for these moduli


// Low-level inline functions


// Computes a * b % m
inline rawtype modmul (rawtype a, rawtype b, rawtype m)
{
    int t;
    rawtype r = 0.0, c = 2251799813685248.0;         // 2^51

    for (t = 52; t; t--)
    {
        if ((r += r) >= m) r -= m;
        if (b >= c)
        {
            if ((r += a) >= m) r -= m;
            b -= c;
        }
        c *= 0.5;
    }

    return r;
}

inline rawtype modadd (rawtype a, rawtype b, rawtype m)
{
    rawtype r;

    r = a + b;
    return (r >= m ? r - m : r);
}

inline rawtype modsub (rawtype a, rawtype b, rawtype m)
{
    rawtype r;

    r = a - b;
    return (r < 0.0 ? r + m : r);
}


#endif  // __RAW_H
